from typing import Callable, List, Optional

import plotly.graph_objects as go
import pandas as pd

from insider.mixins import BaseMixin, MovingIndicatorMixin
from insider.stock import Stock
from insider.constants import MA_N, MD_N, EXPMA_N


class StockInsider(Stock, BaseMixin, MovingIndicatorMixin):
    """Plot daily trading indicators."""

    def __init__(self, code):
        """
        code: Full stock code，(e.g. 'sz002156')，股票完整代码
        """
        super().__init__(code)

    @staticmethod
    def _plot_line(df: pd.DataFrame, head: int, line_name: str):
        if head:
            df = df.tail(head)
        plot_data = go.Scatter(x=df["day"], y=df["close"], name=line_name)
        return plot_data

    def _plot_moving_lines(
        self,
        func: Callable,
        name: str,
        head: int = 90,
        ns: Optional[List] = None,
        verbose: bool = False,
    ):

        plot_data = []
        for n in ns:
            df = func(n=n)
            line_name = name + str(n)
            plot_data.append(self._plot_line(df, head, line_name))

        if verbose:
            df = self._df.copy()
            stock_data = self._plot_stock_data(df, head)
            plot_data.append(stock_data)

        layout = self._set_layout()
        fig = go.Figure(data=plot_data, layout=layout)
        if verbose:
            fig.update_layout(xaxis_rangeslider_visible=False)
        fig.update_layout(title_text=f"{name.upper()} Chart ({self.stock_code})")
        fig.show()

    def plot_ma(
        self, head: int = 90, ns: Optional[List[int]] = None, verbose: bool = False
    ):
        """Plot Moving Average Indicator. 绘出MA曲线

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
            ns: Select which trading lines to plot, default is to plot 5, 10, 20, 60-day lines
            选择曲线的种类，e.g. [5, 10], 默认会绘出5, 10, 20, 60日曲线
            verbose: If to include stock price or not, default is False.
            选择是否将股票价格曲线一起绘出，默认是False，将会只绘出指标曲线。
        """
        if ns is None:
            ns = MA_N

        func = self.ma
        self._plot_moving_lines(func=func, name="ma", head=head, ns=ns, verbose=verbose)

    def plot_md(
        self, head: int = 90, ns: Optional[List[int]] = None, verbose: bool = False
    ):
        """Plot Moving Deviation Indicator. 绘出MD曲线

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
            ns: Select which trading lines to plot, default is to plot 5, 10, 20-day lines
            选择曲线的种类，e.g. [5, 10], 默认会绘出5, 10, 20日曲线
            verbose: If to include stock price or not, default is False.
            选择是否将股票价格曲线一起绘出，默认是False，将会只绘出指标曲线。
        """
        if ns is None:
            ns = MD_N

        func = self.md
        self._plot_moving_lines(func=func, name="md", head=head, ns=ns, verbose=verbose)

    def plot_ema(
        self, head: int = 90, ns: Optional[List[int]] = None, verbose: bool = False
    ):
        """Plot Exponential Moving Average Indicator. 绘出EMA曲线

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。
            ns: Select which trading lines to plot, default is to plot 5, 10, 20, 60-day lines
            选择曲线的种类，e.g. [5, 10, 20, 60], 默认会绘出5, 10, 20, 60日曲线
            verbose: If to include stock price or not, default is False.
            选择是否将股票价格曲线一起绘出，默认是False，将会只绘出指标曲线。
        """
        if ns is None:
            ns = EXPMA_N

        func = self.ema
        self._plot_moving_lines(
            func=func, name="ema", head=head, ns=ns, verbose=verbose
        )

    def plot_macd(self, head: int = 90):
        """Plot MACD Indicator. 绘出MACD曲线

        Parameters:
            head: The recent number of trading days to plot, default is 90, 最近交易日的天数，
            默认90，将会绘出最近90个交易日的曲线。

        A mixed chart will be plotted, including a bar chart to visualize MACD, and line charts
        to visualize DIFF and DEA.
        将会绘出差值柱形图来表示MACD, 以及表示差离值和讯号线的线性图。
        """
        df_macd = self.macd()
        if head:
            df_macd = df_macd.tail(head)

        df_macd.loc[:, "color"] = df_macd["macd"].apply(
            lambda x: "red" if x >= 0 else "green"
        )

        layout = self._set_layout()
        fig = go.Figure(layout=layout)

        fig.add_trace(
            go.Bar(
                x=df_macd["day"],
                y=df_macd["macd"],
                base=0,
                marker_color=df_macd["color"],
                name="MACD",
            )
        )
        fig.add_trace(
            go.Scatter(
                x=df_macd["day"], y=df_macd["dea"], marker_color="orange", name="DEA"
            )
        )
        fig.add_trace(
            go.Scatter(
                x=df_macd["day"], y=df_macd["diff"], marker_color="black", name="DIFF"
            )
        )
        fig.update_layout(title_text=f"MACD Chart ({self.stock_code})")
        fig.show()



from insider.constants import MOVING_COLS


class BaseMixin:
    """Base Mixins used in different Indicator Mixins"""

    def _ma(self, col, n, df=None):
        if df is None:
            df = self._df
        return df[col].rolling(n).mean()

    def _md(self, col, n, df=None):
        if df is None:
            df = self._df
        return df[col].rolling(n).std(ddof=0)

    def _ema(self, col, n, df=None):
        if df is None:
            df = self._df
        return df[col].ewm(ignore_na=False, span=n, min_periods=0, adjust=False).mean()


class MovingIndicatorMixin:
    """Moving Indicator Mixin (移动指标混合)"""

    def ma(self, n=5):
        """Moving Average Calculation (移动平均值计算)"""
        df_ma = self._df.loc[:, MOVING_COLS]
        df_ma.loc[:, "close"] = self._ma(col="close", n=n)
        return df_ma

    def md(self, n=5):
        """Moving Deviation Calculation (移动标准差值计算)"""
        df_md = self._df.loc[:, MOVING_COLS]
        df_md.loc[:, "close"] = self._md(col="close", n=n)
        return df_md

    def ema(self, n=5):
        """Exponential Moving Average Calculation (指数移动平均值计算)"""
        df_ema = self._df.loc[:, MOVING_COLS]
        df_ema.loc[:, "close"] = self._ema(col="close", n=n)
        return df_ema

    def macd(self, n=12, m=26, k=9):
        """Moving Average Convergence Divergence Calculation (平滑异同移动平均计算)"""
        df_macd = self._df.loc[:, MOVING_COLS]
        df_macd.loc[:, "diff"] = self._ema(col="close", n=n) - self._ema(
            col="close", n=m
        )
        df_macd.loc[:, "dea"] = self._ema(col="diff", n=k, df=df_macd)
        df_macd.loc[:, "macd"] = 2 * (df_macd["diff"] - df_macd["dea"])
        return df_macd



# Constants used in Stock class
STOCK_URL = (
    "http://api.finance.ifeng.com/{ktype}/?code={code}&type=last"
)  # URL to fetch stock information
MA_COLS = ["ma5", "ma10", "ma20"]  # columns from dataset to use to plot MA lines
MA_COLORS = ["black", "orange", "red"]  # colors to choose to plot different MA lines
KTYPE_CONVERSION = {
    "D": "akdaily",
    "W": "akweekly",
    "M": "akmonthly",
}  # Conversion for the type of K-lines for URL
KTYPES = ["D", "W", "M"]  # Allowed types of K-lines
DAY_COL = ["day"]  # date column from fetched dataset
NUMERIC_COLUMNS = [
    "open",
    "high",
    "close",
    "low",
    "volumn",
    "price_change",
    "percent_change",
    "ma5",
    "ma10",
    "ma20",
    "v_ma5",
    "v_ma10",
    "v_ma20",
]  # all numeric columns from fetched dataset

# Constants used in Mixins
MOVING_COLS = ["day", "close"]  # cols used for Moving indicator mixin

# Constants used in StockInsider class
MA_N = [5, 10, 20, 60]  # Number of days counted for MA indicator
MD_N = [5, 10, 20]  # Number of days counted for MD indicator
EXPMA_N = [5, 10, 20, 60]  # Number of days counted for EMA indicator



from os import path
from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

__keywords__ = ["chinese stock", "visualization", "insight", "stock analysis", "python"]

setup(
    name="StockInsider",
    description="The Python implementation of an stock analysis and visualization tool.",
    long_description_content_type='text/markdown',
    long_description=long_description,
    version="0.0.1",
    author="Kaiqi Dong",
    url="https://github.com/charlesdong1991/StockInsider",
    author_email="kaiqidong1991@gmail.com",
    licnese="MIT License",
    keywords=__keywords__,
    packages=find_packages(exclude=["test*"]),
    install_requires=["numpy>=1.18.3", "plotly>=4.6.0", "pandas>=1.0.3", "requests>=2.23.0"],
    tests_require=["pytest"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Libraries",
    ],
)
